from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

import itertools
import functools
from functools import reduce
import operator

from xoutil.future.itertools import map as imap, zip as izip

from ..lang import H
from ..lang import sig
from ..lang import t
from ..lang import L
from ..lang import __
from ..lang import caseof
from ..lang import m
from ..lang import p

from .Eq import Eq
from .Ord import Ord
from .Ord import Ordering
from .Num import Num
from .Num import Integral
from .Maybe import Maybe
from .Maybe import Just
from .Maybe import Nothing


@sig(H/ ["a"] >> "a")
def head(xs):
    """``head :: [a] -> a``

    Extract the first element of a list, which must be non-empty.

    """
    return xs[0]


@sig(H/ ["a"] >> "a")
def last(xs):
    """``last :: [a] -> a``

    Extract the last element of a list, which must be finite and non-empty.

    """
    return xs[-1]


@sig(H/ ["a"] >> ["a"])
def tail(xs):
    """``tail :: [a] -> [a]``

    Extract the elements after the head of a list, which must be non-empty.

    """
    if null(xs):
        raise IndexError("empty list")
    return xs[1:]


@sig(H/ ["a"] >> ["a"])
def init(xs):
    """``init :: [a] -> [a]``

    Return all the elements of a list except the last one. The list must be
    non-empty.

    """
    if null(xs):
        raise IndexError("empty list")
    return xs[:-1]


@sig(H/ ["a"] >> t(Maybe, ("a", ["a"])))
def uncons(xs):
    """``uncons :: [a] -> Maybe (a, [a])``

    Decompose a list into its head and tail. If the list is empty, returns
    Nothing. If the list is non-empty, returns Just((x, xs)), where x is the
    head of the list and xs its tail.

    """
    return Just((head(xs), tail(xs))) if not null(xs) else Nothing


@sig(H/ ["a"] >> bool)
def null(xs):
    """``null :: [a] -> bool``

    Test whether the structure is empty.

    """
    return ~(caseof(xs)
                | m(m.y ^ m.ys) >> False
                | m(m.ys)       >> True)


@sig(H/ ["a"] >> int)
def length(xs):
    """``length :: [a] -> int``

    Returns the size/length of a finite structure as an Int. The default
    implementation is optimized for structures that are similar to cons-lists,
    because there is no general way to do better.

    """
    return len(xs)


@sig(H/ (H/ "a" >> "b") >> ["a"] >> ["b"])
def map(f, xs):
    """``map :: (a -> b) -> [a] -> [b]``

    map(f, xs) is the list obtained by applying f to each element of xs

    """
    return L[imap(f, xs)]


@sig(H/ ["a"] >> ["a"])
def reverse(xs):
    """``reverse :: [a] -> [a]``

    reverse(xs) returns the elements of xs in reverse order. xs must be finite.

    """
    return L[(x for x in xs[::-1])]


@sig(H/ "a" >> ["a"] >> ["a"])
def intersperse(x, xs):
    """``intersperse :: a -> [a] -> [a]``

    The intersperse function takes an element and a list and intersperses that
    element between the elements of the list.

    """
    def __intersperse(x, xs):
        for y in init(xs):
            yield y
            yield x
        yield last(xs)

    if null(xs):
        return xs
    return L[__intersperse(x, xs)]


@sig(H/ ["a"] >> [["a"]] >> ["a"])
def intercalate(xs, xss):
    """``intercalate :: [a] -> [[a]] -> [a]``

    intercalate(xs,xss) is equivalent to concat(intersperse(xs,xss)). It
    inserts the list xs in between the lists in xss and concatenates the
    result.

    """
    return concat(intersperse(xs, xss))


@sig(H/ [["a"]] >> [["a"]])
def transpose(xs):
    """``transpose :: [[a]] -> [[a]]``

    The transpose function transposes the rows and columns of its argument.

    """
    return L[(L[x] for x in izip(*xs))]


@sig(H/ ["a"] >> [["a"]])
def subsequences(xs):
    """``subsequences :: [a] -> [[a]]``

    The subsequences function returns the list of all subsequences of the
    argument.

    """
    ret = L[[L[[]]]]
    for r, _ in enumerate(xs):
        ret += L[(L[x] for x in itertools.combinations(xs, r+1))]
    return ret


@sig(H/ ["a"] >> [["a"]])
def permutations(xs):
    """``permutations :: [a] -> [[a]]``

    The permutations function returns the list of all permutations of the
    argument.

    """
    if null(xs):
        return L[[]]
    return L[(L[x] for x in itertools.permutations(xs))]


@sig(H/ (H/ "b" >> "a" >> "b") >> "a" >> ["a"] >> "b")
def foldl(f, z, xs):
    """``foldl :: (b -> a -> b) -> b -> [a] -> b``

    foldl, applied to a binary operator, a starting value (typically the
    left-identity of the operator), and a list, reduces the list using the
    binary operator, from left to right. The list must be finite.

    """
    return reduce(f, xs, z)


@sig(H/ (H/ "b" >> "a" >> "b") >> "a" >> ["a"] >> "b")
def foldl_(f, z, xs):
    """``foldl_ :: (b -> a -> b) -> b -> [a] -> b``

    A strict version of foldl.

    """
    return foldl(f, z, xs)


@sig(H/ (H/ "a" >> "a" >> "a") >> ["a"] >> "a")
def foldl1(f, xs):
    """``foldl1 :: (a -> a -> a) -> [a] -> a``

    foldl1 is a variant of foldl that has no starting value argument, and thus
    must be applied to non-empty lists.

    """
    return foldl(f, xs[0], xs[1:])


@sig(H/ (H/ "a" >> "a" >> "a") >> ["a"] >> "a")
def foldl1_(f, xs):
    """``foldl1_ :: (a -> a -> a) -> [a] -> a``

    A strict version of foldl1

    """
    return foldl1(f, xs[0], xs[1:])


@sig(H/ (H/ "a" >> "b" >> "b") >> "a" >> ["a"] >> "b")
def foldr(f, z, xs):
    """``foldr :: (a -> b -> b) -> b -> [a] -> b``

    foldr, applied to a binary operator, a starting value (typically the
    right-identity of the operator), and a list, reduces the list using the
    binary operator, from right to left

    """
    return ~(caseof(xs)
                | m(L[[]])     >> z
                | m(m.a ^ m.b) >> f(p.a, foldr(f, z, p.b)))


@sig(H/ (H/ "a" >> "a" >> "a") >> ["a"] >> "a")
def foldr1(f, xs):
    """``foldr1 :: (a -> a -> a) -> [a] -> a``

    foldr1 is a variant of foldr that has no starting value argument, and thus
    must be applied to non-empty lists.

    """
    return foldr(f, xs[0], xs[1:])


@sig(H/ [["a"]] >> ["a"])
def concat(xss):
    """``concat :: [[a]] -> [a]``

    Concatenate a list of lists.

    """
    def __concat(xss):
        for xs in xss:
            for x in xs:
                yield x
    return L[__concat(xss)]


@sig(H/ (H/ "a" >> ["b"]) >> ["a"] >> ["b"])
def concatMap(f, xs):
    """``concatMap :: (a -> [b]) -> [a] -> [b]``

    Map a function over a list and concatenate the results.

    """
    return concat(map(f, xs))


@sig(H/ [bool] >> bool)
def and_(xs):
    """``and_ :: [Bool] -> Bool``

    and returns the conjunction of a Boolean list. For the result to be True,
    the list must be finite; False, however, results from a False value at a
    finite index of a finite or infinite list.

    """
    return False not in xs


@sig(H/ [bool] >> bool)
def or_(xs):
    """``or_ :: [Bool] -> Bool``

    or returns the disjunction of a Boolean list. For the result to be False,
    the list must be finite; True, however, results from a True value at a
    finite index of a finite or infinite list.

    """
    return True in xs


@sig(H/ (H/ "a" >> bool) >> ["a"] >> bool)
def any(p, xs):
    """``any :: (a -> Bool) -> [a] -> Bool``

    Applied to a predicate and a list, any determines if any element of the
    list satisfies the predicate. For the result to be False, the list must be
    finite; True, however, results from a True value for the predicate applied
    to an element at a finite index of a finite or infinite list.

    """
    return True in ((p(x) for x in xs))


@sig(H/ (H/ "a" >> bool) >> ["a"] >> bool)
def all(p, xs):
    """``all :: (a -> Bool) -> [a] -> Bool``

    Applied to a predicate and a list, all determines if all elements of the
    list satisfy the predicate. For the result to be True, the list must be
    finite; False, however, results from a False value for the predicate
    applied to an element at a finite index of a finite or infinite list.

    """
    return False not in ((p(x) for x in xs))


@sig(H[(Num, "a")]/ ["a"] >> "a")
def sum(xs):
    """``sum :: Num a => [a] -> a``

    The sum function computes the sum of a finite list of numbers.

    """
    return functools.reduce(operator.add, xs, 0)


@sig(H[(Num, "a")]/ ["a"] >> "a")
def product(xs):
    """``product :: Num a => [a] -> a``

    The product function computes the product of a finite list of numbers.

    """
    return functools.reduce(operator.mul, xs, 1)


@sig(H[(Ord, "a")]/ ["a"] >> "a")
def minimum(xs):
    """``minimum :: Ord a => [a] -> a``

    minimum returns the minimum value from a list, which must be non-empty,
    finite, and of an ordered type. It is a special case of minimumBy, which
    allows the programmer to supply their own comparison function.

    """
    return min(xs)


@sig(H[(Ord, "a")]/ ["a"] >> "a")
def maximum(xs):
    """``maximum :: Ord a => [a] -> a``

    maximum returns the maximum value from a list, which must be non-empty,
    finite, and of an ordered type. It is a special case of maximumBy, which
    allows the programmer to supply their own comparison function.

    """
    return max(xs)


@sig(H/ (H/ "b" >> "a" >> "b") >> "b" >> ["a"] >> ["b"])
def scanl(f, z, xs):
    """``scanl :: (b -> a -> b) -> b -> [a] -> [b]``

    scanl is similar to foldl, but returns a list of successive reduced values
    from the left

    """
    raise NotImplementedError()


@sig(H/ (H/ "a" >> "a" >> "a") >> ["a"] >> ["a"])
def scanl1(f, xs):
    """``scanl1 :: (a -> a -> a) -> [a] -> [a]``

    scanl1 is a variant of scanl that has no starting value argument

    """
    raise NotImplementedError()


@sig(H/ (H/ "a" >> "a" >> "b") >> "b" >> ["a"] >> ["b"])
def scanr(f, z, xs):
    """``scanr :: (a -> b -> b) -> b -> [a] -> [b]``

    scanr is the right-to-left dual of scanl.

    """
    raise NotImplementedError()


@sig(H/ (H/ "a" >> "a" >> "a") >> ["a"] >> ["a"])
def scanr1(f, xs):
    """``scanr1 :: (a -> a -> a) -> [a] -> [a]``

    scanr1 is a variant of scanr that has no starting value argument.

    """
    raise NotImplementedError()


@sig(H/ (H/ "a" >> "x" >> ("a", "y")) >> "a" >> ["x"] >> ("a", ["y"]))
def mapAccumL(xs):
    """``mapAccumL :: (a -> x -> (a, y)) -> a -> [x] -> (a, [y])``

    The mapAccumL function behaves like a combination of map and foldl; it
    applies a function to each element of a list, passing an accumulating
    parameter from left to right, and returning a final value of this
    accumulator together with the new list.

    """
    raise NotImplementedError()


@sig(H/ (H/ "a" >> "x" >> ("a", "y")) >> "a" >> ["x"] >> ("a", ["y"]))
def mapAccumR(xs):
    """``mapAccumR :: (acc -> x -> (acc, y)) -> acc -> [x] -> (acc, [y])``

    The mapAccumR function behaves like a combination of map and foldr; it
    applies a function to each element of a list, passing an accumulating
    parameter from right to left, and returning a final value of this
    accumulator together with the new list.

    """
    raise NotImplementedError()


@sig(H/ (H/ "a" >> "a") >> "a" >> ["a"])
def iterate(f, x):
    """``iterate :: (a -> a) -> a -> [a]``

    iterate(f, x) returns an infinite List of repeated applications of f to x:
    iterate(f, x) == [x, f(x), f(f(x)), ...]

    """
    def __iterate(f, x):
        while True:
            yield x
            x = f(x)
    return L[__iterate(f, x)]


@sig(H/ "a" >> ["a"])
def repeat(x):
    """``repeat :: a -> [a]``

    repeat(x) is an infinite list, with x the value of every element.

    """
    def __repeat(x):
        while True:
            yield x
    return L[__repeat(x)]


@sig(H/ int >> "a" >> ["a"])
def replicate(n, x):
    """``replicate :: Int -> a -> [a]``

    replicate(n, x) is a list of length n with x the value of every element.

    """
    def __replicate(n, x):
        for _ in range(n):
            yield x
    return L[__replicate(n, x)]


@sig(H/ ["a"] >> ["a"])
def cycle(x):
    """``cycle :: [a] -> [a]``

    cycle ties a finite list into a circular one, or equivalently, the infinite
    repetition of the original list. It is the identity on infinite lists.

    """
    def __cycle(x):
        while True:
            for i in x:
                yield i
    return L[__cycle(x)]


@sig(H/ (H/ "b" >> t(Maybe, ("a", "b"))) >> "b" >> ["a"])
def unfoldr(f, x):
    """``unfoldr :: (b -> Maybe (a, b)) -> b -> [a]``

    The unfoldr function is a dual to foldr: while foldr reduces a list to a
    summary value, unfoldr builds a list from a seed value. The function takes
    the element and returns Nothing if it is done producing the list or
    returns Just (a,b), in which case, a is prepended to the list and b is
    used as the next element in a recursive call

    """
    y = f(x)
    if y == Nothing:
        return L[[]]
    return y[0][0] ^ unfoldr(f, y[0][1])


@sig(H/ int >> ["a"] >> ["a"])
def take(n, xs):
    """``take :: Int -> [a] -> [a]``

    take(n), applied to a list xs, returns the prefix of xs of length n, or xs
    itself if n > length xs

    """
    return xs[:n]


@sig(H/ int >> ["a"] >> ["a"])
def drop(n, xs):
    """``drop :: Int -> [a] -> [a]``

    drop(n, xs) returns the suffix of xs after the first n elements, or [] if n >
    length xs

    """
    return xs[n:]


@sig(H/ int >> ["a"] >> (["a"], ["a"]))
def splitAt(n, xs):
    """``splitAt :: Int -> [a] -> ([a], [a])``

    splitAt(n, xs) returns a tuple where first element is xs prefix of length n
    and second element is the remainder of the list

    """
    return (xs[:n], xs[n:])


@sig(H/ (H/ "a" >> bool) >> ["a"] >> ["a"])
def takeWhile(p, xs):
    """``takeWhile :: (a -> Bool) -> [a] -> [a]``

    takeWhile, applied to a predicate p and a list xs, returns the longest
    prefix (possibly empty) of xs of elements that satisfy p

    """
    return L[itertools.takewhile(p, xs)]


@sig(H/ (H/ "a" >> bool) >> ["a"] >> ["a"])
def dropWhile(p, xs):
    """``dropWhile :: (a -> Bool) -> [a] -> [a]``

    dropWhile(p, xs) returns the suffix remaining after takeWhile(p, xs)

    """
    return L[itertools.dropwhile(p, xs)]


@sig(H/ (H/ "a" >> bool) >> ["a"] >> ["a"])
def dropWhileEnd(p, xs):
    """``dropWhileEnd :: (a -> Bool) -> [a] -> [a]``

    The dropWhileEnd function drops the largest suffix of a list in which the
    given predicate holds for all elements.

    """
    raise NotImplementedError()


@sig(H/ (H/ "a" >> bool)  >> ["a"] >> (["a"], ["a"]))
def span(p, xs):
    """``span :: (a -> Bool) -> [a] -> ([a], [a])``

    span, applied to a predicate p and a list xs, returns a tuple where first
    element is longest prefix (possibly empty) of xs of elements that satisfy p
    and second element is the remainder of the list

    """
    front = takeWhile(p, xs)
    rest = xs[len(front):]
    return front, rest


@sig(H/ (H/ "a" >> bool)  >> ["a"] >> (["a"], ["a"]))
def break_(p, xs):
    """``break :: (a -> Bool) -> [a] -> ([a], [a])``

    break, applied to a predicate p and a list xs, returns a tuple where first
    element is longest prefix (possibly empty) of xs of elements that do not
    satisfy p and second element is the remainder of the list

    """
    inv = (lambda x: not p(x)) ** (H/ "a" >> bool)
    return span(inv, xs)


@sig(H[(Eq, "a")]/ ["a"] >> ["a"] >> t(Maybe, ["a"]))
def stripPrefix(xs, ys):
    """``stripPrefix :: Eq a => [a] -> [a] -> Maybe [a]``

    The stripPrefix function drops the given prefix from a list. It returns
    Nothing if the list did not start with the prefix given, or Just the list
    after the prefix, if it does.

    """
    if isPrefixOf(xs, ys):
        return Just(ys[len(xs)])
    return Nothing


@sig(H[(Eq, "a")]/ ["a"] >> [["a"]])
def group(xs):
    """``group :: Eq a => [a] -> [[a]]``

    The group function takes a list and returns a list of lists such that the
    concatenation of the result is equal to the argument. Moreover, each
    sublist in the result contains only equal elements.
    It is a special case of groupBy, which allows the programmer to supply
    their own equality test.

    """
    return groupBy(xs, (__==__))


@sig(H/ ["a"] >> [["a"]])
def inits(xs):
    """``inits :: [a] -> [[a]]``

    The inits function returns all initial segments of the argument, shortest
    first.

    """
    if null(xs):
        return L[[xs]]
    return L[[L[[]]]] + L[(xs[:n+1] for n, _ in enumerate(xs))]


@sig(H/ ["a"] >> [["a"]])
def tails(xs):
    """``tails :: [a] -> [[a]]``

    The tails function returns all final segments of the argument, longest
    first.

    """
    if null(xs):
        return L[[L[[]]]]
    return L[(xs[n:] for n, _ in enumerate(xs))] + L[[L[[]]]]


@sig(H[(Eq, "a")]/ ["a"] >> ["a"] >> bool)
def isPrefixOf(xs, ys):
    """``isPrefixOf :: Eq a => [a] -> [a] -> Bool``

    The isPrefixOf function takes two lists and returns True iff the first list
    is a prefix of the second.

    """
    return xs == ys[:len(xs)]


@sig(H[(Eq, "a")]/ ["a"] >> ["a"] >> bool)
def isSuffixOf(xs, ys):
    """``isSuffixOf :: Eq a => [a] -> [a] -> Bool``

    The isSuffixOf function takes two lists and returns True iff the first list
    is a suffix of the second. The second list must be finite.

    """
    return xs == ys[-len(xs):]


@sig(H[(Eq, "a")]/ ["a"] >> ["a"] >> bool)
def isInfixOf(xs, ys):
    """``isInfixOf :: Eq a => [a] -> [a] -> Bool``

    The isInfixOf function takes two lists and returns True iff the first list
    is contained, wholly and intact, anywhere within the second.

    """
    return any(isPrefixOf(xs), tails(ys))


@sig(H[(Eq, "a")]/ ["a"] >> ["a"] >> bool)
def isSubsequenceOf(x, y):
    """``isSubsequenceOf :: Eq a => [a] -> [a] -> Bool``

    The isSubsequenceOf function takes two lists and returns True if the first
    list is a subsequence of the second list.

    isSubsequenceOf(x, y) is equivalent to elem(x, subsequences(y))

    """
    return elem(x, subsequences(y))


@sig(H[(Eq, "a")]/ "a" >> ["a"] >> bool)
def elem(x, xs):
    """``elem :: Eq a => a -> [a] -> Bool``

    elem is the list membership predicate, elem(x, xs). For the result to be
    False, the list must be finite; True, however, results from an element
    equal to x found at a finite index of a finite or infinite list.

    """
    return x in xs


@sig(H[(Eq, "a")]/ "a" >> ["a"] >> bool)
def notElem(x, xs):
    """``notElem :: Eq a => a -> [a] -> Bool``

    notElem is the negation of `elem`:func:.

    """
    return not elem(x, xs)


@sig(H[(Eq, "a")]/ "a" >> [("a", "b")] >> t(Maybe, "b"))
def lookup(key, assocs):
    """``lookup :: Eq a => a -> [(a, b)] -> Maybe b``

    lookup(key, assocs) looks up a key in an association list.

    """
    for k, value in assocs:
        if k == key:
            return Just(value)
    return Nothing


@sig(H/ (H/ "a" >> bool) >> ["a"] >> t(Maybe, "a"))
def find(p, xs):
    """``find :: (a -> Bool) -> [a] -> Maybe a``

    The find function takes a predicate and a structure and returns the
    leftmost element of the structure matching the predicate, or Nothing if
    there is no such element.

    """
    for x in xs:
        if p(x):
            return Just(x)
    return Nothing


@sig(H/ (H/ "a" >> bool) >> ["a"] >> ["a"])
def filter(f, xs):
    """``filter :: (a -> Bool) -> [a] -> [a]``

    filter, applied to a predicate and a list, returns the list of those
    elements that satisfy the predicate

    """
    return L[itertools.ifilter(f, xs)]


@sig(H/ (H/ "a" >> bool) >> ["a"] >> (["a"], ["a"]))
def partition(f, xs):
    """``partition :: (a -> Bool) -> [a] -> ([a], [a])``

    The partition function takes a predicate a list and returns the pair of
    lists of elements which do and do not satisfy the predicate.

    """
    yes, no = [], []
    for item in xs:
        if f(item):
            yes.append(item)
        else:
            no.append(item)
    return L[yes], L[no]


@sig(H[(Eq, "a")]/ "a" >> ["a"] >> t(Maybe, int))
def elemIndex(x, xs):
    """``elemIndex :: Eq a => a -> [a] -> Maybe Int``

    The elemIndex function returns the index of the first element in the given
    list which is equal (by ==) to the query element, or Nothing if there is no
    such element.

    """
    for i, a in enumerate(xs):
        if a == x:
            return Just(i)
    return Nothing


@sig(H[(Eq, "a")]/ "a" >> ["a"] >> [int])
def elemIndices(x, xs):
    """``elemIndices :: Eq a => a -> [a] -> [Int]``

    The elemIndices function extends elemIndex, by returning the indices of all
    elements equal to the query element, in ascending order.

    """
    return L[(i for i, a in enumerate(xs) if a == x)]


@sig(H/ (H/ "a" >> bool) >> ["a"] >> t(Maybe, int))
def findIndex(f, xs):
    """``findIndex :: (a -> Bool) -> [a] -> Maybe Int``

    The findIndex function takes a predicate and a list and returns the index
    of the first element in the list satisfying the predicate, or Nothing if
    there is no such element.

    """
    for i, x in enumerate(xs):
        if f(x):
            return Just(i)
    return Nothing


@sig(H/ (H/ "a" >> bool) >> ["a"] >> [int])
def findIndicies(f, xs):
    """``findIndices :: (a -> Bool) -> [a] -> [Int]``

    The findIndices function extends findIndex, by returning the indices of all
    elements satisfying the predicate, in ascending order.

    """
    return L[(i for i, x in enumerate(xs) if f(x))]


@sig(H/ ["a"] >> ["b"] >> [("a", "b")])
def zip(xs, ys):
    """``zip :: [a] -> [b] -> [(a, b)]``

    zip takes two lists and returns a list of corresponding pairs. If one input
    list is short, excess elements of the longer list are discarded.

    """
    return L[izip(xs, ys)]


@sig(H/ ["a"] >> ["b"] >> ["c"] >> [("a", "b", "c")])
def zip3(a, b, c):
    """``zip3 :: [a] -> [b] -> [c] -> [(a, b, c)]``

    zip3 takes three lists and returns a list of triples, analogous to zip.

    """
    return L[izip(a, b, c)]


@sig(H/ ["a"] >> ["b"] >> ["c"] >> ["d"] >> [("a", "b", "c", "d")])
def zip4(a, b, c, d):
    """``zip4 :: [a] -> [b] -> [c] -> [d] -> [(a, b, c, d)]``

    The zip4 function takes four lists and returns a list of quadruples,
    analogous to zip.

    """
    return L[izip(a, b, c, d)]


@sig(H/ ["a"] >> ["b"] >> ["c"] >> ["d"] >> ["e"] >>
        [("a", "b", "c", "d", "e")])
def zip5(a, b, c, d, e):
    """``zip5 :: [a] -> [b] -> [c] -> [d] -> [e] -> [(a, b, c, d, e)]``

    The zip5 function takes five lists and returns a list of five-tuples,
    analogous to zip.

    """
    return L[izip(a, b, c, d, e)]


@sig(H/ ["a"] >> ["b"] >> ["c"] >> ["d"] >> ["e"] >> ["f"] >>
        [("a", "b", "c", "d", "e", "f")])
def zip6(a, b, c, d, e, f):
    """``zip6 :: [a] -> [b] -> [c] -> [d] -> [e] -> [f] -> [(a, b, c, d, e, f)]``

    The zip6 function takes six lists and returns a list of six-tuples,
    analogous to zip.

    """
    return L[izip(a, b, c, d, e, f)]


@sig(H/ ["a"] >> ["b"] >> ["c"] >> ["d"] >> ["e"] >> ["f"] >> ["g"] >>
        [("a", "b", "c", "d", "e", "f", "g")])
def zip7(a, b, c, d, e, f, g):
    """``zip7 :: [a] -> [b] -> [c] -> [d] -> [e] -> [f] -> [g] -> [(a, b, c, d, e, f, g)]``

    The zip7 function takes seven lists and returns a list of seven-tuples,
    analogous to zip.

    """
    return L[izip(a, b, c, d, e, f, g)]


@sig(H/ (H/ "a" >> "b" >> "c") >> ["a"] >> ["b"] >> ["c"])
def zipWith(fn, xs, ys):
    """``zipWith :: (a -> b -> c) -> [a] -> [b] -> [c]``

    zipWith generalises zip by zipping with the function given as the first
    argument, instead of a tupling function. For example, zipWith (+) is
    applied to two lists to produce the list of corresponding sums.

    """
    return L[(fn(*s) for s in zip(xs, ys))]


@sig(H/ (H/ "a" >> "b" >> "c" >> "d") >> ["a"] >> ["b"] >> ["c"] >> ["d"])
def zipWith3(fn, a, b, c):
    """``zipWith3 :: (a -> b -> c -> d) -> [a] -> [b] -> [c] -> [d]``

    The zipWith3 function takes a function which combines three elements, as
    well as three lists and returns a list of their point-wise combination,
    analogous to zipWith.

    """
    return L[(fn(*s) for s in zip3(a, b, c))]


@sig(H/ (H/ "a" >> "b" >> "c" >> "d" >> "e") >>
        ["a"] >> ["b"] >> ["c"] >> ["d"] >> ["e"])
def zipWith4(fn, a, b, c, d):
    """``zipWith4 :: (a -> b -> c -> d -> e) -> [a] -> [b] -> [c] -> [d] -> [e]``

    The zipWith4 function takes a function which combines four elements, as
    well as four lists and returns a list of their point-wise combination,
    analogous to zipWith.

    """
    return L[(fn(*s) for s in zip4(a, b, c, d))]


@sig(H/ (H/ "a" >> "b" >> "c" >> "d" >> "e" >> "f") >>
        ["a"] >> ["b"] >> ["c"] >> ["d"] >> ["e"] >> ["f"])
def zipWith5(fn, a, b, c, d, e):
    """``zipWith5 :: (a -> b -> c -> d -> e -> f) -> [a] -> [b] -> [c] -> [d] -> [e] -> [f]``

    The zipWith5 function takes a function which combines five elements, as
    well as five lists and returns a list of their point-wise combination,
    analogous to zipWith.

    """
    return L[(fn(*s) for s in zip5(a, b, c, d, e))]


@sig(H/ (H/ "a" >> "b" >> "c" >> "d" >> "e" >> "f" >> "g") >>
        ["a"] >> ["b"] >> ["c"] >> ["d"] >> ["e"] >> ["f"] >> ["g"])
def zipWith6(fn, a, b, c, d, e, f):
    """``zipWith6 :: (a -> b -> c -> d -> e -> f -> g) -> [a] -> [b] -> [c] -> [d] -> [e] -> [f] -> [g]``

    The zipWith6 function takes a function which combines six elements, as well
    as six lists and returns a list of their point-wise combination, analogous
    to zipWith.

    """
    return L[(fn(*s) for s in zip6(a, b, c, d, e, f))]


@sig(H/ (H/ "a" >> "b" >> "c" >> "d" >> "e" >> "f" >> "g" >> "h") >>
        ["a"] >> ["b"] >> ["c"] >> ["d"] >> ["e"] >> ["f"] >> ["g"] >> ["h"])
def zipWith7(fn, a, b, c, d, e, f):
    """``zipWith7 :: (a -> b -> c -> d -> e -> f -> g -> h) -> [a] -> [b] -> [c] -> [d] -> [e] -> [f] -> [g] -> [h]``

    The zipWith7 function takes a function which combines seven elements, as
    well as seven lists and returns a list of their point-wise combination,
    analogous to zipWith.

    """
    return L[(fn(*s) for s in zip7(a, b, c, d, e, f))]


@sig(H/ [("a", "b")] >> (["a"], ["b"]))
def unzip(xs):
    """``unzip :: [(a, b)] -> ([a], [b])``

    unzip transforms a list of pairs into a list of first components and a
    list of second components.

    """
    a = L[(i[0] for i in xs)]
    b = L[(i[1] for i in xs)]
    return a, b


@sig(H/ [("a", "b", "c")] >> (["a"], ["b"], ["c"]))
def unzip3(xs):
    """``unzip3 :: [(a, b, c)] -> ([a], [b], [c])``

    The unzip3 function takes a list of triples and returns three lists,
    analogous to unzip.

    """
    a = L[(i[0] for i in xs)]
    b = L[(i[1] for i in xs)]
    c = L[(i[2] for i in xs)]
    return a, b, c


@sig(H/ [("a", "b", "c", "d")] >> (["a"], ["b"], ["c"], ["d"]))
def unzip4(xs):
    """``unzip4 :: [(a, b, c, d)] -> ([a], [b], [c], [d])``

    The unzip4 function takes a list of quadruples and returns four lists,
    analogous to unzip.

    """
    a = L[(i[0] for i in xs)]
    b = L[(i[1] for i in xs)]
    c = L[(i[2] for i in xs)]
    d = L[(i[3] for i in xs)]
    return a, b, c, d


@sig(H/ [("a", "b", "c", "d", "e")] >> (["a"], ["b"], ["c"], ["d"], ["e"]))
def unzip5(xs):
    """``unzip5 :: [(a, b, c, d, e)] -> ([a], [b], [c], [d], [e])``

    The unzip5 function takes a list of five-tuples and returns five lists,
    analogous to unzip.

    """
    a = L[(i[0] for i in xs)]
    b = L[(i[1] for i in xs)]
    c = L[(i[2] for i in xs)]
    d = L[(i[3] for i in xs)]
    e = L[(i[4] for i in xs)]
    return a, b, c, d, e


@sig(H/ [("a", "b", "c", "d", "e", "f")]
        >> (["a"], ["b"], ["c"], ["d"], ["e"], ["f"]))
def unzip6(xs):
    """``unzip6 :: [(a, b, c, d, e, f)] -> ([a], [b], [c], [d], [e], [f])``

    The unzip6 function takes a list of six-tuples and returns six lists,
    analogous to unzip.

    """
    a = L[(i[0] for i in xs)]
    b = L[(i[1] for i in xs)]
    c = L[(i[2] for i in xs)]
    d = L[(i[3] for i in xs)]
    e = L[(i[4] for i in xs)]
    f = L[(i[5] for i in xs)]
    return a, b, c, d, e, f


@sig(H/ [("a", "b", "c", "d", "e", "f", "g")]
        >> (["a"], ["b"], ["c"], ["d"], ["e"], ["f"], ["g"]))
def unzip7(xs):
    """``unzip7 :: [(a, b, c, d, e, f, g)] -> ([a], [b], [c], [d], [e], [f], [g])``

    The unzip7 function takes a list of seven-tuples and returns seven lists,
    analogous to unzip.

    """
    a = L[(i[0] for i in xs)]
    b = L[(i[1] for i in xs)]
    c = L[(i[2] for i in xs)]
    d = L[(i[3] for i in xs)]
    e = L[(i[4] for i in xs)]
    f = L[(i[5] for i in xs)]
    g = L[(i[6] for i in xs)]
    return a, b, c, d, e, f, g


from .String import lines  # noqa
from .String import words  # noqa
from .String import unlines  # noqa
from .String import unwords  # noqa


@sig(H[(Eq, "a")]/ ["a"] >> ["a"])
def nub(xs):
    """``nub :: Eq a => [a] -> [a]``

    The nub function removes duplicate elements from a list. In particular, it
    keeps only the first occurrence of each element. (The name nub means
    essence.) It is a special case of nubBy, which allows the programmer to
    supply their own equality test.

    """
    return L[(i for i in set(xs))]


@sig(H[(Eq, "a")]/ "a" >> ["a"] >> ["a"])
def delete(x, xs):
    """
    ``delete :: Eq a => a -> [a] -> [a]``

    delete(x) removes the first occurrence of x from its list argument.

    It is a special case of deleteBy, which allows the programmer to supply
    their own equality test.
    """
    raise NotImplementedError()


@sig(H[(Eq, "a")]/ ["a"] >> ["a"] >> ["a"])
def diff(xs, ys):
    """``diff :: :: Eq a => [a] -> [a] -> [a]``

    List difference.

    """
    raise NotImplementedError()


@sig(H[(Eq, "a")]/ ["a"] >> ["a"] >> ["a"])
def union(xs, ys):
    """``union :: Eq a => [a] -> [a] -> [a]``

    The union function returns the list union of the two lists.

    Duplicates, and elements of the first list, are removed from the the second
    list, but if the first list contains duplicates, so will the result. It is
    a special case of unionBy, which allows the programmer to supply their own
    equality test.

    """
    raise NotImplementedError()


@sig(H[(Eq, "a")]/ ["a"] >> ["a"] >> ["a"])
def intersect(xs, ys):
    """``intersect :: Eq a => [a] -> [a] -> [a]``

    The intersect function takes the list intersection of two lists.  It is a
    special case of intersectBy, which allows the programmer to supply their
    own equality test. If the element is found in both the first and the second
    list, the element from the first list will be used.

    """
    raise NotImplementedError()


@sig(H[(Ord, "a")]/ ["a"] >> ["a"])
def sort(xs):
    """``sort :: Ord a => [a] -> [a]``

    The sort function implements a stable sorting algorithm. It is a special
    case of sortBy, which allows the programmer to supply their own comparison
    function.

    Note: Current implementation is not lazy

    """
    return L[sorted(xs)]


@sig(H[(Ord, "b")]/ (H/ "a" >> "b") >> ["a"] >> ["a"])
def sortOn(f, xs):
    """``sortOn :: Ord b => (a -> b) -> [a] -> [a]``

    Sort a list by comparing the results of a key function applied to each
    element.

    Note: Current implementation is not lazy

    """
    raise NotImplementedError()


@sig(H[(Ord, "a")]/ "a" >> ["a"] >> ["a"])
def insert(x, xs):
    """``insert :: Ord a => a -> [a] -> [a]``

    The insert function takes an element and a list and inserts the element
    into the list at the first position where it is less than or equal to the
    next element. In particular, if the list is sorted before the call, the
    result will also be sorted.

    """
    def __insert(x, xs):
        for i in xs:
            if i > x:
                yield x
            yield i
    return L[__insert(x, xs)]


@sig(H/ (H/ "a" >> "a" >> bool) >> ["a"] >> ["a"])
def nubBy(f, xs):
    """``nubBy :: (a -> a -> Bool) -> [a] -> [a]``

    The nubBy function behaves just like nub, except it uses a user-supplied
    equality predicate instead of the overloaded == function.

    """
    raise NotImplementedError()


@sig(H/ (H/ "a" >> "a" >> bool) >> "a" >> ["a"] >> ["a"])
def deleteBy(f, xs):
    """``deleteBy :: (a -> a -> Bool) -> a -> [a] -> [a]``

    The deleteBy function behaves like delete, but takes a user-supplied
    equality predicate.

    """
    raise NotImplementedError()


@sig(H/ (H/ "a" >> "a" >> bool) >> ["a"] >> ["a"] >> ["a"])
def deleteFirstBy(f, xs, ys):
    """``deleteFirstsBy :: (a -> a -> Bool) -> [a] -> [a] -> [a]``

    The deleteFirstsBy function takes a predicate and two lists and returns the
    first list with the first occurrence of each element of the second list
    removed.

    """
    raise NotImplementedError()


@sig(H/ (H/ "a" >> "a" >> bool) >> ["a"] >> ["a"] >> ["a"])
def unionBy(f, xs, ys):
    """``unionBy :: (a -> a -> Bool) -> [a] -> [a] -> [a]``

    The unionBy function is the non-overloaded version of union.

    """
    raise NotImplementedError()


@sig(H/ (H/ "a" >> "a" >> bool) >> ["a"] >> ["a"] >> ["a"])
def intersectBy(f, xs, ys):
    """``intersectBy :: (a -> a -> Bool) -> [a] -> [a] -> [a]``

    The intersectBy function is the non-overloaded version of intersect.

    """
    raise NotImplementedError()


@sig(H/ (H/ "a" >> "a" >> bool) >> ["a"] >> [["a"]])
def groupBy(f, xs):
    """``groupBy :: (a -> a -> Bool) -> [a] -> [[a]]``

    The groupBy function is the non-overloaded version of group.

    """
    return L[(L[i[1]] for i in itertools.groupby(xs, keyfunc=f))]


@sig(H/ (H/ "a" >> "a" >> Ordering) >> ["a"] >> ["a"])
def sortBy(f, xs):
    """``sortBy :: (a -> a -> Ordering) -> [a] -> [a]``

    The sortBy function is the non-overloaded version of sort.

    """
    raise NotImplementedError()


@sig(H/ (H/ "a" >> "a" >> Ordering) >> "a" >> ["a"] >> ["a"])
def insertBy(f, xs):
    """``insertBy :: (a -> a -> Ordering) -> a -> [a] -> [a]``

    The non-overloaded version of insert.

    """
    raise NotImplementedError()


@sig(H/ (H/ "a" >> "a" >> Ordering) >> ["a"] >> "a")
def maximumBy(f, xs):
    """``maximumBy :: (a -> a -> Ordering) -> [a] -> a``

    The maximumBy function takes a comparison function and a list and returns
    the greatest element of the list by the comparison function. The list must
    be finite and non-empty.

    """
    return max(xs, key=f)


@sig(H/ (H/ "a" >> "a" >> Ordering) >> ["a"] >> "a")
def minimumBy(f, xs):
    """``minimumBy :: (a -> a -> Ordering) -> [a] -> a``

    The minimumBy function takes a comparison function and a list and returns
    the least element of the list by the comparison function. The list must be
    finite and non-empty.

    """
    return min(xs, key=f)


@sig(H[(Num, "i")]/ ["a"] >> "i")
def genericLength(xs):
    """``genericLength :: Num i => [a] -> i``

    The genericLength function is an overloaded version of length. In
    particular, instead of returning an Int, it returns any type which is an
    instance of Num. It is, however, less efficient than length.

    """
    raise NotImplementedError()


@sig(H[Integral, "i"]/ "i" >> ["a"] >> ["a"])
def genericTake(n, xs):
    """``genericTake :: Integral i => i -> [a] -> [a]``

    The genericTake function is an overloaded version of take, which accepts
    any Integral value as the number of elements to take.

    """
    raise NotImplementedError()


@sig(H[Integral, "i"]/ "i" >> ["a"] >> ["a"])
def genericDrop(n, xs):
    """``genericDrop :: Integral i => i -> [a] -> [a]``

    The genericDrop function is an overloaded version of drop, which accepts
    any Integral value as the number of elements to drop.

    """
    raise NotImplementedError()


@sig(H[Integral, "i"]/ "i" >> ["a"] >> (["a"], ["a"]))
def genericSplitAt(n, xs):
    """``genericSplitAt :: Integral i => i -> [a] -> ([a], [a])``

    The genericSplitAt function is an overloaded version of splitAt, which
    accepts any Integral value as the position at which to split.

    """
    raise NotImplementedError()


@sig(H[Integral, "i"]/ ["a"] >> "i" >> ["a"])
def genericIndex(xs, i):
    """``genericIndex :: Integral i => [a] -> i -> a``

    The genericIndex function is an overloaded version of !!, which accepts any
    Integral value as the index.

    """
    raise NotImplementedError()


@sig(H[Integral, "i"]/ "i" >> ["a"] >> ["a"])
def genericReplicate(i, a):
    """``genericReplicate :: Integral i => i -> a -> [a]``

    The genericReplicate function is an overloaded version of replicate, which
    accepts any Integral value as the number of repetitions to make.

    """
    raise NotImplementedError()



from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from xoutil.future.itertools import map as imap

from ..lang import TypedFunc
from ..lang import Typeclass
from ..lang import is_builtin
from ..lang import build_instance
from ..lang import List
from ..lang import L
from ..lang import H
from ..lang import sig
from ..lang import t
from ..lang import instance


class Functor(Typeclass):
    """
    The Functor class is used for types that can be mapped over. Instances of
    Functor should satisfy the following laws::

        fmap(id)  ==  id
        fmap(f * g)  ==  fmap(f * (fmap g))

    Attributes:

    - ``fmap``
    - ``__rmul__``

    Minimal complete definition:

    - ``fmap``

    """
    @classmethod
    def make_instance(typeclass, cls, fmap):
        fmap = fmap ** \
            (H[(Functor, "f")]/ (H/ "a" >> "b") >> t("f", "a") >> t("f", "b"))
        if not is_builtin(cls):
            cls.__rmul__ = lambda x, f: fmap(f, x)
        build_instance(Functor, cls, {"fmap": fmap})


@sig(H[(Functor, "f")]/ (H/ "a" >> "b") >> t("f", "a") >> t("f", "b"))
def fmap(f, x):
    return Functor[x].fmap(f, x)


instance(Functor, List).where(
    fmap = lambda fn, lst: L[imap(fn, iter(lst))]
)

instance(Functor, TypedFunc).where(
    fmap = TypedFunc.__mul__
)



from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

import inspect
import operator
import sys
from collections import deque, defaultdict

from xoutil.objects import get_first_of

from .type_system import typeof
from .type_system import Typeclass
from .type_system import TypedFunc
from .type_system import TypeSignature
from .type_system import TypeSignatureHKT
from .type_system import ADT
from .type_system import build_ADT
from .type_system import build_sig
from .type_system import make_fn_type
from .type_system import PatternMatchBind
from .type_system import PatternMatchListBind
from .type_system import pattern_match
from .type_system import Undefined
from .type_system import PyFunc


__magic_methods__ = ["__%s__" % s for s in set((
    "len", "getitem", "setitem", "delitem", "iter", "reversed", "contains",
    "missing", "delattr", "call", "enter", "exit", "eq", "ne", "gt", "lt",
    "ge", "le", "pos", "neg", "abs", "invert", "round", "floor", "ceil",
    "trunc", "add", "sub", "mul", "div", "truediv", "floordiv", "mod",
    "divmod", "pow", "lshift", "rshift", "or", "and", "xor", "radd", "rsub",
    "rmul", "rdiv", "rtruediv", "rfloordiv", "rmod", "rdivmod", "rpow",
    "rlshift", "rrshift", "ror", "rand", "rxor", "isub", "imul", "ifloordiv",
    "idiv", "imod", "idivmod", "irpow", "ilshift", "irshift", "ior", "iand",
    "ixor", "nonzero"))]


def replace_magic_methods(cls, fn):
    """
    Replace the magic method of a class with some function or method.

    Args:
        cls: The class to modify
        fn: The function to replace cls's magic methods with
    """
    for attr in __magic_methods__:
        setattr(cls, attr, fn)


class Syntax(object):
    """
    Base class for new syntactic constructs. All of the new "syntax" elements
    of Hask inherit from this class.

    By default, a piece of syntax will raise a syntax error with a standard
    error message if the syntax object is used with a Python builtin operator.

    Subclasses may override these methods to define what syntax is valid for
    those objects.

    """
    def __init__(self, err_msg):
        self.__syntax_err_msg = err_msg
        self.invalid_syntax = SyntaxError(self.__syntax_err_msg)

    def __raise(self):
        raise self.invalid_syntax

    __syntaxerr__ = lambda s, *a: s.__raise()


replace_magic_methods(Syntax, Syntax.__syntaxerr__)


class instance(Syntax):
    """
    Special syntax for defining typeclass instances.

    Example usage::

        instance(Functor, Maybe).where(
            fmap = ...
        )

    """
    def __init__(self, typecls, cls):
        if not (inspect.isclass(typecls) and issubclass(typecls, Typeclass)):
            raise TypeError("%s is not a typeclass" % typecls)
        self.typeclass = typecls
        self.cls = cls

    def where(self, **kwargs):
        self.typeclass.make_instance(self.cls, **kwargs)


class __constraints__(Syntax):
    """H/ creates a new function type signature.

    Examples::

        (H/ int >> int >> int)
        (H/ (H/ "a" >> "b" >> "c") >> "b" >> "a" >> "c")
        (H/ func >> set >> set)
        (H/ (H/ "a" >> "b") >> ["a"] >> ["b"])
        (H[(Eq, "a")]/ "a" >> ["a"] >> bool)
        (H/ int >> int >> t(Maybe, int))
        (H/ int >> None)

    See `sig`:class: for more information on type signature decorators.

    """
    def __init__(self, constraints=None):
        self.constraints = defaultdict(lambda: [])
        if constraints:
            # multiple typeclass constraints
            if isinstance(constraints[0], tuple):
                for con in constraints:
                    self.__add_constraint(con)
            # only one typeclass constraint
            else:
                self.__add_constraint(constraints)
        super(__constraints__, self).__init__("Syntax error in type signature")

    def __add_constraint(self, con):
        if len(con) != 2 or not isinstance(con, tuple):
            raise SyntaxError("Invalid typeclass constraint: %s" % str(con))

        if not isinstance(con[1], str):
            raise SyntaxError("%s is not a type variable" % con[1])

        if not (inspect.isclass(con[0]) and issubclass(con[0], Typeclass)):
            raise SyntaxError("%s is not a typeclass" % con[0])

        self.constraints[con[1]].append(con[0])

    def __getitem__(self, constraints):
        return __constraints__(constraints)

    def __div__(self, arg):
        return __signature__((), self.constraints).__rshift__(arg)

    def __truediv__(self, arg):
        return self.__div__(arg)


class __signature__(Syntax):
    """
    Class that represents a (complete or incomplete) type signature.
    """
    def __init__(self, args, constraints):
        self.sig = TypeSignature(args, constraints)
        super(__signature__, self).__init__("Syntax error in type signature")

    def __rshift__(self, arg):
        arg = arg.sig if isinstance(arg, __signature__) else arg
        return __signature__(self.sig.args + (arg,), self.sig.constraints)

    def __rpow__(self, fn):
        return sig(self)(fn)


H = __constraints__()
func = PyFunc


class sig(Syntax):
    """Decorator to convert a Python function into a statically typed function
    (`~hask.lang.type_system.TypedFunc`:class: object).

    TypedFuncs are automagically curried, and polymorphic type arguments will
    be inferred by the type system.

    Usage::

        @sig(H/ int >> int >> int )
        def add(x, y):
            return x + y

        @sig(H[(Show, "a")]/ >> "a" >> str)
        def to_str(x):
            return str(x)

    """
    def __init__(self, signature):
        super(self.__class__, self).__init__("Syntax error in type signature")

        if not isinstance(signature, __signature__):
            msg = "Signature expected in sig(); found %s" % signature
            raise SyntaxError(msg)

        elif len(signature.sig.args) < 2:
            raise SyntaxError("Not enough type arguments in signature")

        self.sig = signature.sig

    def __call__(self, fn):
        fn_args = build_sig(self.sig)
        fn_type = make_fn_type(fn_args)
        return TypedFunc(fn, fn_args, fn_type)


def t(type_constructor, *params):
    '''Helper to instantiate `~hask.lang.type_system.TypeSignatureHKT`:class:.

    '''
    if inspect.isclass(type_constructor) and \
       issubclass(type_constructor, ADT) and \
       len(type_constructor.__params__) != len(params):
            raise TypeError("Incorrect number of type parameters to %s" %
                            type_constructor.__name__)

    params = [p.sig if isinstance(p, __signature__) else p for p in params]
    return TypeSignatureHKT(type_constructor, params)


def typify(fn, hkt=None):
    """Convert an untyped Python function to a TypeFunc.

    :param fn: The function to wrap

    :param hkt: A higher-kinded type wrapped in a closure (e.g.,
               ``lambda x: t(Maybe, x)``)

    :returns: A `~hask.lang.type_system.TypedFunc`:class: object with a
        polymorphic type (e.g. ``a -> b -> c``, etc) with the same number of
        arguments as `fn`.  If `hkt` is supplied, the return type will be the
        supplied HKT parameterized by a type variable.

    Example usage::

        @typify(hkt=lambda x: t(Maybe, x))
        def add(x, y):
            return x + y

    """
    code = get_first_of(fn, '__code__', 'fn_code')
    args = [chr(i) for i in range(97, 98 + code.co_argcount)]
    if hkt is not None:
        args[-1] = hkt(args[-1])
    return sig(__signature__(args, []))


class __undefined__(Undefined):
    """
    Undefined value with special syntactic powers. Whenever you try to use one
    if its magic methods, it returns undefined. Used to prevent overzealous
    evaluation in pattern matching.

    Its type unifies with any other type.
    """
    pass


replace_magic_methods(__undefined__, lambda *a: __undefined__())
undefined = __undefined__()


# Constructs for pattern matching.
# Note that the approach implemented here uses lots of global state and is
# pretty much the opposite of "functional" or "thread-safe."

class IncompletePatternError(Exception):
    pass


class MatchStackFrame(object):
    """One stack frame for pattern matching bound variable stack"""
    def __init__(self, value):
        self.value = value
        self.cache = {}
        self.matched = False


class MatchStack(object):
    """Stack for storing locally bound variables from matches"""
    __stack__ = deque()

    @classmethod
    def push(cls, value):
        """Push a new frame onto the stack, representing a new case expr"""
        cls.__stack__.append(MatchStackFrame(value))

    @classmethod
    def pop(cls):
        """Pop the current frame off the stack"""
        cls.__stack__.pop()

    @classmethod
    def get_frame(cls):
        """Access the current frame"""
        return cls.__stack__[-1]

    @classmethod
    def get_name(cls, name):
        """Lookup a variable name in the current frame"""
        if cls.get_frame().matched:
            return undefined
        return cls.get_frame().cache.get(name, undefined)


class __var_bind__(Syntax):
    """
    ``m.*`` binds a local variable while pattern matching.

    For example usage, see `caseof`:class:.

    """
    def __getattr__(self, name):
        return __pattern_bind__(name)

    def __call__(self, pattern):
        is_match, env = pattern_match(MatchStack.get_frame().value, pattern)
        if is_match and not MatchStack.get_frame().matched:
            MatchStack.get_frame().cache = env
        return __match_test__(is_match)


class __var_access__(Syntax):
    """``p.*`` accesses a local variable bound during pattern matching.

    For example usage, see `caseof`:class:.

    """
    def __getattr__(self, name):
        return MatchStack.get_name(name)


m = __var_bind__("Syntax error in pattern match")
p = __var_access__("Syntax error in pattern match")


class __pattern_bind_list__(Syntax, PatternMatchListBind):
    """
    Class that represents a pattern designed to match an iterable, consisting
    of a head (one element) and a tail (zero to many elements).
    """
    def __init__(self, head, tail):
        self.head = [head]
        self.tail = tail
        super(__pattern_bind_list__, self).__init__("Syntax error in match")

    def __rxor__(self, head):
        self.head.insert(0, head)
        return self


class __pattern_bind__(Syntax, PatternMatchBind):
    """
    Class that represents a pattern designed to match any value and bind it to
    a name.
    """
    def __init__(self, name):
        self.name = name
        super(__pattern_bind__, self).__init__("Syntax error in match")

    def __rxor__(self, cell):
        return __pattern_bind_list__(cell, self)

    def __xor__(self, other):
        if isinstance(other, __pattern_bind_list__):
            return other.__rxor__(self)

        elif isinstance(other, __pattern_bind__):
            return __pattern_bind_list__(self, other)

        raise self.invalid_syntax


class __match_line__(Syntax):
    """One line of a caseof expression, i.e.: ``m( ... ) >> return_value``

    """
    def __init__(self, is_match, return_value):
        self.is_match = is_match
        self.return_value = return_value


class __match_test__(Syntax):
    """The pattern part of one caseof line, i.e.: ``m( ... )``

    """
    def __init__(self, is_match):
        self.is_match = is_match

    def __rshift__(self, value):
        MatchStack.get_frame().cache = {}
        return __match_line__(self.is_match, value)


class __unmatched_case__(Syntax):
    """A caseof expression in mid-evaluation, when zero or more lines have been
    tested, but before a match has been found.

    """
    def __or__(self, line):
        if line.is_match:
            MatchStack.get_frame().matched = True
            return __matched_case__(line.return_value)
        return self

    def __invert__(self):
        value = MatchStack.get_frame().value
        MatchStack.pop()
        raise IncompletePatternError(value)


class __matched_case__(Syntax):
    """A caseof expression in mid-evaluation, when one or more lines have been
    tested and after a match has been found.

    """
    def __init__(self, return_value):
        self.value = return_value

    def __or__(self, line):
        return self

    def __invert__(self):
        MatchStack.pop()
        return self.value


class caseof(__unmatched_case__):
    """Pattern matching can be used to deconstruct lists and ADTs, and is a very
    useful control flow tool.

    Usage::

        ~(caseof(value_to_match)
            | m(pattern_1) >> return_value_1
            | m(pattern_2) >> return_value_2
            | m(pattern_3) >> return_value_3)

    Example usage::

        def fib(x):
            return ~(caseof(x)
                        | m(0)   >> 1
                        | m(1)   >> 1
                        | m(m.n) >> fib(p.n - 1) + fib(p.n - 2))

    """
    def __init__(self, value):
        if isinstance(value, Undefined):
            return
        MatchStack.push(value)


# ADT creation syntax ("data" expressions)
# "data"/type constructor half of the expression

class __data__(Syntax):
    r"""`data` is part of Hask's special syntax for defining ADTs.

    Example usage:

        >>> from hask import data, d, deriving, Read, Show, Eq, Ord

        >>> Maybe, Nothing, Just =\
        ...     data.Maybe("a") == d.Nothing | d.Just("a") & \
        ...     deriving(Read, Show, Eq, Ord)

    """
    def __init__(self):
        super(__data__, self).__init__("Syntax error in `data`")

    def __getattr__(self, value):
        if not value[0].isupper():
            raise SyntaxError("Type constructor name must be capitalized")
        return __new_tcon_enum__(value)


class __new_tcon__(Syntax):
    """
    Base class for Syntax classes related to creating new type constructors.
    """
    def __init__(self, name, args=()):
        self.name = name
        self.args = args
        super(__new_tcon__, self).__init__("Syntax error in `data`")

    def __eq__(self, d):
        # one data constructor, zero or more derived typeclasses
        if isinstance(d, __new_dcon__):
            return build_ADT(self.name, self.args, [(d.name, d.args)], d.classes)

        # one or more data constructors, zero or more derived typeclasses
        elif isinstance(d, __new_dcons_deriving__):
            return build_ADT(self.name, self.args, d.dcons, d.classes)

        raise self.invalid_syntax


class __new_tcon_enum__(__new_tcon__):
    """
    This class represents a `data` statement in mid evaluation; it represents
    the part of the expression that builds the type constructor, before type
    parameters have been added.

    Examples::

       data.Either
       data.Ordering

    """
    def __call__(self, *typeargs):
        if len(typeargs) < 1:
            msg = "Missing type args in statement: `data.%s()`" % self.name
            raise SyntaxError(msg)

        # make sure all type params are strings
        if not all((type(arg) == str for arg in typeargs)):
            raise SyntaxError("Type parameters must be strings")

        # make sure all type params are letters only
        is_letters = lambda xs: all(x.islower() for x in xs)
        if not all((is_letters(arg) for arg in typeargs)):
            raise SyntaxError("Type parameters must be lowercase letters")

        # all type parameters must have unique names
        if len(typeargs) != len(set(typeargs)):
            raise SyntaxError("Type parameters are not unique")

        return __new_tcon_hkt__(self.name, typeargs)


class __new_tcon_hkt__(__new_tcon__):
    """
    This class represents a `data` statement in mid evaluation; it represents
    the part of the expression that builds the type constructor, after type
    parameters have been added.

    Examples::

        data.Maybe("a")
        data.Either("a", "b")

    """
    pass


# "d"/data constructor half of the expression
class __d__(Syntax):
    """`d` is part of hask's special syntax for defining algebraic data types.

    See `data`:obj: for more information.

    """
    def __init__(self):
        super(__d__, self).__init__("Syntax error in `d`")

    def __getattr__(self, value):
        if not value[0].isupper():
            raise SyntaxError("Data constructor name must be capitalized")
        return __new_dcon_enum__(value)


class __new_dcon__(Syntax):
    """
    Base class for Syntax objects that handle data constructor creation syntax
    within a `data` statment (`d.*`).
    """
    def __init__(self, dcon_name, args=(), classes=()):
        self.name = dcon_name
        self.args = args
        self.classes = classes
        super(__new_dcon__, self).__init__("Syntax error in `d`")


class __new_dcon_params__(__new_dcon__):
    """
    This class represents a `data` statement in mid evaluation; it represents
    the part of the expression that builds a data constructor, after type
    parameters have been added.

    Examples::

        d.Just("a")
        d.Foo(int, "a", "b", str)

    """
    def __and__(self, derive_exp):
        if not isinstance(derive_exp, deriving):
            raise self.invalid_syntax
        return __new_dcon_deriving__(self.name, self.args, derive_exp.classes)

    def __or__(self, dcon):
        if isinstance(dcon, __new_dcon__):
            constructors = ((self.name, self.args), (dcon.name, dcon.args))

            if isinstance(dcon, __new_dcon_deriving__):
                return __new_dcons_deriving__(constructors, dcon.classes)
            return __new_dcons__(constructors)

        raise self.invalid_syntax


class __new_dcon_deriving__(__new_dcon__):
    """
    This class represents a `data` statement in mid evaluation; it represents
    the part of the expression that builds a data constructor (with or without type
    parameters) and adds derived typeclasses.

    Examples::

        d.Just("a") & deriving(Show, Eq, Ord)
        d.Bar & deriving(Eq)

    """
    pass


class __new_dcon_enum__(__new_dcon_params__):
    """
    This class represents a `data` statement in mid evaluation; it represents
    the part of the expression that builds a data constructor, after type
    parameters have been added.

    Examples::

        d.Just
        d.Foo

    """
    def __call__(self, *typeargs):
        return __new_dcon_params__(self.name, typeargs)


class __new_dcons_deriving__(Syntax):
    """
    This class represents a `data` statement in mid evaluation; it represents
    the part of the expression that builds data constructors (with or without type
    parameters) and adds derived typeclasses.

    Examples::

        d.Nothing | d.Just("a") & deriving(Show, Eq, Ord)
        d.Foo(int, "a", "b", str) | d.Bar & deriving(Eq)

    """
    def __init__(self, data_consts, classes=()):
        self.dcons = data_consts
        self.classes = classes
        super(__new_dcons_deriving__, self).__init__("Syntax error in `d`")


class __new_dcons__(__new_dcons_deriving__):
    """
    This class represents a `data` statement in mid evaluation; it represents
    the part of the expression that builds data constructors (with or without type
    parameters), with no derived typeclasses.

    Examples::

       d.Foo(int, "a", "b", str) | d.Bar

    """
    def __init__(self, data_consts):
        super(__new_dcons__, self).__init__(data_consts)

    def __or__(self, new_dcon):
        if isinstance(new_dcon, __new_dcon__):
            constructor = ((new_dcon.name, new_dcon.args),)

            if isinstance(new_dcon, __new_dcon_deriving__):
                return __new_dcons_deriving__(self.dcons + constructor,
                                              new_dcon.classes)
            return __new_dcons__(self.dcons + constructor)
        raise self.invalid_syntax


data = __data__()
d = __d__()


class deriving(Syntax):
    """`deriving` is part of hask's special syntax for defining algebraic data
    types.

    See `data`:class: for more information.

    """
    def __init__(self, *tclasses):
        for tclass in tclasses:
            if not issubclass(tclass, Typeclass):
                raise TypeError("Cannot derive non-typeclass %s" % tclass)
        self.classes = tclasses
        super(deriving, self).__init__("Syntax error in `deriving`")


class __section__(Syntax):
    """The class of the ``__`` object.

    This is Hask's special syntax for operator sections: a placeholder for
    arguments (operands).

    Example usage:

        >>> from __future__ import division
        >>> from hask.lang import __

        >>> (__+1)(5)
        6

        >>> (6//__) * (__-1) % 4
        2

        >>> (__**__)(2, 10)
        1024

    Operators supported::

        + - * / // ** >> << | & ^ == != > >= < <=

    """
    def __init__(self, syntax_err_msg):
        super(__section__, self).__init__(syntax_err_msg)

    @staticmethod
    def __make_section(fn):
        """
        Create an operator section from a binary operator.
        """
        def section_wrapper(self, y):
            # double section, e.g. (__+__)
            if isinstance(y, __section__):
                @sig(H/ "a" >> "b" >> "c")
                def double_section(a, b):
                    return fn(a, b)
                return double_section

            # single section, e.g. (__+1) or (1+__)
            @sig(H/ "a" >> "b")
            def section(a):
                return fn(a, y)
            return section
        return section_wrapper

    # left section, e.g. (__+1)
    __wrap = __make_section.__func__

    # right section, e.g. (1+__)
    __flip = lambda f: lambda x, y: f(y, x)

    __add__ = __wrap(operator.add)
    __sub__ = __wrap(operator.sub)
    __mul__ = __wrap(operator.mul)
    __truediv__ = __wrap(operator.truediv)
    __floordiv__ = __wrap(operator.floordiv)
    __mod__ = __wrap(operator.mod)
    __divmod__ = __wrap(divmod)
    __pow__ = __wrap(operator.pow)
    __lshift__ = __wrap(operator.lshift)
    __rshift__ = __wrap(operator.rshift)
    __or__ = __wrap(operator.or_)
    __and__ = __wrap(operator.and_)
    __xor__ = __wrap(operator.xor)

    __eq__ = __wrap(operator.eq)
    __ne__ = __wrap(operator.ne)
    __gt__ = __wrap(operator.gt)
    __lt__ = __wrap(operator.lt)
    __ge__ = __wrap(operator.ge)
    __le__ = __wrap(operator.le)

    __radd__ = __wrap(__flip(operator.add))
    __rsub__ = __wrap(__flip(operator.sub))
    __rmul__ = __wrap(__flip(operator.mul))
    __rtruediv__ = __wrap(__flip(operator.truediv))
    __rfloordiv__ = __wrap(__flip(operator.floordiv))
    __rmod__ = __wrap(__flip(operator.mod))
    __rdivmod__ = __wrap(__flip(divmod))
    __rpow__ = __wrap(__flip(operator.pow))
    __rlshift__ = __wrap(__flip(operator.lshift))
    __rrshift__ = __wrap(__flip(operator.rshift))
    __ror__ = __wrap(__flip(operator.or_))
    __rand__ = __wrap(__flip(operator.and_))
    __rxor__ = __wrap(__flip(operator.xor))

    if sys.version[0] == '2':
        __div__ = __wrap(operator.div)
        __rdiv__ = __wrap(__flip(operator.div))


__ = __section__("Error in section")


# Guards! Guards!

# Unlike pattern matching, this approach is completely stateless and
# thread-safe. However, it has the pretty undesireable property that it cannot
# be used with recursive functions.


class NoGuardMatchException(Exception):
    pass


class __guard_test__(Syntax):
    """A case in a guard.

    ``c`` creates a new condition that can be used in a guard expression.

    ``otherwise`` is a guard condition that always evaluates to True.

    Usage::

        ~(guard(<expr to test>)
            | c(<test_fn_1>) >> <return_value_1>
            | c(<test_fn_2>) >> <return_value_2>
            | otherwise      >> <return_value_3>
        )

    See `guard`:class: for more details.

    """
    def __init__(self, fn):
        if not callable(fn):
            raise ValueError("Guard condition must be callable")
        self.__test = fn
        super(__guard_test__, self).__init__("Syntax error in guard condition")

    def __rshift__(self, value):
        if isinstance(value, __guard_test__) or \
           isinstance(value, __guard_conditional__) or \
           isinstance(value, __guard_base__):
            raise self.invalid_syntax
        return __guard_conditional__(self.__test, value)


class __guard_conditional__(Syntax):
    """One line of a guard expression.

    Consists of:

    1) a condition (a test function wrapped in c and a value to be returned
       if that condition is satisfied).

    2) a return value, which will be returned if the condition evaluates
       to True

    See `guard`:class: for more details.

    """
    def __init__(self, fn, return_value):
        self.check = fn
        self.return_value = return_value
        msg = "Syntax error in guard condition"
        super(__guard_conditional__, self).__init__(msg)


class __guard_base__(Syntax):
    """Superclass for the classes __unmatched_guard__ and __matched_guard__
    below, which represent the internal state of a guard expression as it is
    being evaluated.

    See `guard`:class: for more details.

    """
    def __init__(self, value):
        self.value = value
        super(__guard_base__, self).__init__("Syntax error in guard")


class __unmatched_guard__(__guard_base__):
    """Object that represents the state of a guard expression in mid-evaluation,
    before one of the conditions in the expression has been satisfied.

    See `guard`:class: for more details.

    """
    def __or__(self, cond):
        # Consume the next line of the guard expression

        if isinstance(cond, __guard_test__):
            raise SyntaxError("Guard expression is missing return value")

        elif not isinstance(cond, __guard_conditional__):
            raise SyntaxError("Guard condition expected, got %s" % cond)

        # If the condition is satisfied, change the evaluation state to
        # __matched_guard__, setting the return value to the value provided on
        # the current line
        elif cond.check(self.value):
            return __matched_guard__(cond.return_value)

        # If the condition is not satisfied, continue on with the next line,
        # still in __unmatched_guard__ state with the return value not set
        return __unmatched_guard__(self.value)

    def __invert__(self):
        raise NoGuardMatchException("No match found in guard(%s)" % self.value)


class __matched_guard__(__guard_base__):
    """Object that represents the state of a guard expression in mid-evaluation,
    after one of the conditions in the expression has been satisfied.

    See `guard`:class: for more details.

    """
    def __or__(self, cond):
        # Consume the next line of the guard expression
        # Since a condition has already been satisfied, we can ignore the rest
        # of the lines in the guard expression
        if isinstance(cond, __guard_conditional__):
            return self
        raise self.invalid_syntax

    def __invert__(self):
        return self.value


class guard(__unmatched_guard__):
    """Special syntax for guard expression.

    Usage::

        ~(guard(<expr to test>)
            | c(<test_fn_1>) >> <return_value_1>
            | c(<test_fn_2>) >> <return_value_2>
            | otherwise      >> <return_value_3>
        )

    Examples::

        ~(guard(8)
             | c(lambda x: x < 5) >> "less than 5"
             | c(lambda x: x < 9) >> "less than 9"
             | otherwise          >> "unsure"
        )

        # Using guards with sections. See help(__) for information on sections.
        ~(guard(20)
            | c(__ > 10)  >> 20
            | c(__ == 10) >> 10
            | c(__ > 5)   >> 5
            | otherwise   >> 0)

    :param value: the value being tested in the guard expression

    :returns: the return value corresponding to the first matching condition

    :raises: NoGuardMatchException (if no match is found)

    """
    def __invert__(self):
        raise self.invalid_syntax


c = __guard_test__
otherwise = c(lambda _: True)


# REPL tools (:q, :t, :i)


def _q(status=None):
    """
    Shorthand for sys.exit() or exit() with no arguments. Equivalent to :q in
    Haskell. Should only be used in the REPL.

    Usage:

    >>> _q()
    """
    if status is None:
        exit()
    exit(status)


def _t(obj):
    """Returns a string representing the type of an object, including
    higher-kinded types and ADTs.

    Equivalent to ``:t`` in Haskell.  Meant to be used in the REPL, but might
    also be useful for debugging.

    :param obj: the object to inspect

    :returns: A string representation of the type

    Usage:

        >>> from hask import _t

        >>> _t(1)
        'int'

        >>> _t(Just("hello world"))
        '(Maybe str)'

    """
    return str(typeof(obj))


def _i(obj):
    """Show information about an object.

    Equivalent to ``:i`` in Haskell or ``help(obj)`` in Python.  Should only
    be used in the REPL.

    :param obj: the object to inspect

    Usage::

        >>> _i(Just("hello world"))
        ...

        >>> _i(Either)
        ...

    """
    help(obj)



from ..lang import H
from ..lang import sig


@sig(H/ str >> bool)
def isControl(s):
    """``isControl :: str -> bool``

    Selects control characters, which are the non-printing characters of the
    Latin-1 subset of Unicode.

    """
    raise NotImplementedError


@sig(H/ str >> bool)
def isSpace(s):
    r"""``isSpace :: str -> bool``

    Returns True for any Unicode space character, and the control characters
    \t, \n, \r, \f, \v.

    """
    return s in " \t\n\r\f\v"


@sig(H/ str >> bool)
def isLower(s):
    """``isLower :: str -> bool``

    Selects lower-case alphabetic Unicode characters (letters).

    """
    return s.lower() == s


@sig(H/ str >> bool)
def isUpper(s):
    """``isUpper :: str -> bool``

    Selects upper-case or title-case alphabetic Unicode characters (letters).
    Title case is used by a small number of letter ligatures like the
    single-character form of Lj.

    """
    return s.upper() == s


@sig(H/ str >> bool)
def isAlpha(s):
    """``isAlpha :: str -> bool``

    Selects alphabetic Unicode characters (lower-case, upper-case and
    title-case letters, plus letters of caseless scripts and modifiers
    letters). This function is equivalent to isLetter.

    """
    raise NotImplementedError


@sig(H/ str >> bool)
def isAlphaNum(s):
    """``isAlphaNum :: str -> bool``

    Selects alphabetic or numeric digit Unicode characters.

    Note that numeric digits outside the ASCII range are selected by this
    function but not by isDigit. Such digits may be part of identifiers but are
    not used by the printer and reader to represent numbers.

    """
    raise NotImplementedError


@sig(H/ str >> bool)
def isPrint(s):
    """``isPrint :: str -> bool``

    Selects printable Unicode characters (letters, numbers, marks, punctuation,
    symbols and spaces).

    """
    raise NotImplementedError


@sig(H/ str >> bool)
def isDigit(s):
    """``isDigit :: str -> bool``

    Selects ASCII digits, i.e. '0'..'9'.

    """
    return s in "0123456789"


@sig(H/ str >> bool)
def isOctDigit(s):
    """``isOctDigit :: str -> bool``

    Selects ASCII octal digits, i.e. '0'..'7'.

    """
    return s in "01234567"


@sig(H/ str >> bool)
def isHexDigit(s):
    """``isHexDigit :: str -> bool``

    Selects ASCII hexadecimal digits, i.e. '0'..'9', 'a'..'f', 'A'..'F'.

    """
    return s in "0123456789abcdefABCDEF"


@sig(H/ str >> bool)
def isLetter(s):
    """``isLetter :: str -> bool``

    Selects alphabetic Unicode characters (lower-case, upper-case and
    title-case letters, plus letters of caseless scripts and modifiers
    letters). This function is equivalent to isAlpha.

    """
    return isAlpha(s)


@sig(H/ str >> bool)
def isMark(s):
    """``isMark :: str -> bool``

    Selects Unicode mark characters, e.g. accents and the like, which combine
    with preceding letters.

    """
    raise NotImplementedError


@sig(H/ str >> bool)
def isNumber(s):
    """``isNumber :: str -> bool``

    Selects Unicode numeric characters, including digits from various scripts,
    Roman numerals, etc.

    """
    raise NotImplementedError


@sig(H/ str >> bool)
def isPunctuation(s):
    """``isPunctuation :: str -> bool``

    Selects Unicode punctuation characters, including various kinds of
    connectors, brackets and quotes.

    """
    raise NotImplementedError


@sig(H/ str >> bool)
def isSymbol(s):
    """``isSymbol :: str -> bool``

    Selects Unicode symbol characters, including mathematical and currency
    symbols.

    """
    raise NotImplementedError


@sig(H/ str >> bool)
def isSeparator(s):
    """``isSeparator :: str -> bool``

    Selects Unicode space and separator characters.

    """
    raise NotImplementedError


@sig(H/ str >> bool)
def isAscii(s):
    """``isAscii :: str -> bool``

    Selects the first 128 characters of the Unicode character set,
    corresponding to the ASCII character set.

    """
    return ord(s) < 128


@sig(H/ str >> bool)
def isLatin1(s):
    """``isLatin1 :: str -> bool``

    Selects the first 256 characters of the Unicode character set,
    corresponding to the ISO 8859-1 (Latin-1) character set.

    """
    return ord(s) < 256


@sig(H/ str >> bool)
def isAsciiUpper(s):
    """``isAsciiUpper :: str -> bool``

    Selects ASCII upper-case letters, i.e. characters satisfying both isAscii
    and isUpper.

    """
    return isAscii(s) and isUpper(s)


@sig(H/ str >> bool)
def isAsciiLower(s):
    """``isAsciiLower :: str -> bool``

    Selects ASCII lower-case letters, i.e. characters satisfying both isAscii
    and isLower.

    """
    return isAscii(s) and isLower(s)


@sig(H/ str >> str)
def toUpper(s):
    """``toUpper :: str -> str``

    Convert a letter to the corresponding upper-case letter, if any. Any other
    character is returned unchanged.

    """
    return s.upper()


@sig(H/ str >> str)
def toLower(s):
    """``toLower :: str -> str``

    Convert a letter to the corresponding lower-case letter, if any. Any other
    character is returned unchanged.

    """
    return s.lower()


@sig(H/ str >> str)
def toTitle(s):
    """``toTitle :: str -> str``

    Convert a letter to the corresponding title-case or upper-case letter, if
    any. (Title case differs from upper case only for a small number of
    ligature letters.) Any other character is returned unchanged.

    """
    return toUpper(s)


@sig(H/ str >> int)
def digitToInt(s):
    """``digitToInt :: str -> int``

    Convert a single digit Char to the corresponding Int. This function fails
    unless its argument satisfies isHexDigit, but recognises both upper and
    lower-case hexadecimal digits (i.e. '0'..'9', 'a'..'f', 'A'..'F').

    """
    if s not in "0123456789abcdefABCDEF":
        raise ValueError("not a digit %s" % s)
    return "0123456789abcdef".index(s.lower())


@sig(H/ int >> str)
def intToDigit(s):
    """``intToDigit :: int -> str``

    Convert an Int in the range 0..15 to the corresponding single digit Char.
    This function fails on other inputs, and generates lower-case hexadecimal
    digits.

    """
    if s > 15 or s < 0:
        raise ValueError("not a digit %s" % s)
    return str(s) if s < 10 else "abcdef"[s-10]


chr = chr ** (H/ int >> str)
ord = ord ** (H/ str >> int)



from setuptools import setup

des = "Haskell language features and standard library ported to Python"
setup(
    name='hask3',
    version='0.1.0',
    description=des,
    long_description=open('README.md').read(),
    author='Bill Murphy',
    author_email='billpmurphy92@gmail.com',
    url='https://github.com/billpmurphy/hask',
    packages=['hask', 'hask.lang', 'hask.Python', 'hask.Data',
              'hask.Control'],
    package_data={'': ['LICENSE', 'README.md']},
    include_package_data=True,
    install_requires=[
        'xoutil>=1.9.4,<2',
    ],
    license=open('LICENSE').read(),
    zip_safe=False,
    classifiers=(
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
    ),
)



from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from ..lang import H
from ..lang import sig
from ..lang import L


@sig(H/ str >> [str])
def lines(string):
    """``lines :: String -> [String]``

    lines breaks a string up into a list of strings at newline characters.
    The resulting strings do not contain newlines.

    """
    return L[[]] if not string else L[string.split("\n")]


@sig(H/ str >> [str])
def words(string):
    """``words :: String -> [String]``

    words breaks a string up into a list of words, which were delimited by
    white space.

    """
    return L[[]] if string == "" else L[string.split(" ")]


@sig(H/ [str] >> str)
def unlines(strings):
    """``lines :: [String] -> String``

    unlines is an inverse operation to lines. It joins lines, after appending a
    terminating newline to each.

    """
    return "\n".join(strings)


@sig(H/ [str] >> str)
def unwords(strings):
    """``unwords :: [String] -> String``

    unwords is an inverse operation to words. It joins words with separating
    spaces.

    """
    return " ".join(strings)



from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from ..lang import Typeclass
from ..lang import build_instance
from ..lang import H
from ..lang import sig


class Monoid(Typeclass):
    """The class of monoids (types with an associative binary operation that has
    an identity)

    Attributes:

    - ``mempty``
    - ``mappend``
    - ``mconcat``

    Minimal complete definition:

    - ``mempty``
    - ``mappend``
    - ``mconcat``

    """
    @classmethod
    def make_instance(typeclass, cls, mempty, mappend, mconcat):
        attrs = {"mempty": mempty, "mappend": mappend, "mconcat": mconcat}
        build_instance(Monoid, cls, attrs)


@sig(H[(Monoid, "m")]/ "m" >> "m" >> "m")
def mappend(x, y):
    """``mappend :: a -> a -> a``

    An associative operation

    """
    return Monoid[x].mappend(x, y)


@sig(H[(Monoid, "m")]/ ["m"] >> "m")
def mconcat(m):
    """``mconcat :: [a] -> a``

    Fold a list using the monoid.

    """
    raise NotImplementedError
